import inspect
from collections.abc import Callable, Mapping
from pathlib import Path
from typing import Any, Literal

import autoregistry
from rich.style import Style
from rich.text import Text

_func = autoregistry.Registry(prefix="_func_")


def func(obj: Callable, *, style: Literal["short", "long"] = "short") -> Text:
    return _func[style](obj)


@_func
def _func_short(obj: Callable) -> Text:
    obj = inspect.unwrap(obj)
    text = Text()
    name: str = _get_name(obj)
    source_file: Path | None = _get_source_file(obj)
    lineno: int = _get_source_lineno(obj)
    if source_file and lineno:
        text.append(
            f"{name}()",
            style=Style(link=f"{source_file.as_uri()}#{lineno}"),
        )
    else:
        text.append(f"{name}()")
    return text


@_func
def _func_long(obj: Callable) -> Text:
    obj = inspect.unwrap(obj)
    text = Text()
    module: str = _get_module(obj)
    qualname: str = _get_qualname(obj)
    source_file: Path | None = _get_source_file(obj)
    lineno: int = _get_source_lineno(obj)
    if source_file:
        text.append(module, style=Style(link=source_file.as_uri()))
        text.append(".")
        text.append(
            f"{qualname}(...)",
            style=Style(link=f"{source_file.as_uri()}#{lineno}"),
        )
    else:
        text.append(f"{module}.{qualname}(...)")
    return text


_pretty_func = func


def call(func: Callable, args: tuple, kwargs: Mapping) -> Text:  # noqa: ARG001
    # TODO: add `args` and `kwargs`
    return _pretty_func(func)


def _get_module(obj: Any) -> str:
    return getattr(obj, "__module__", "unknown")


def _get_name(obj: Any) -> str:
    return getattr(obj, "__name__", "<unknown>")


def _get_qualname(obj: Any) -> str:
    return getattr(obj, "__qualname__", "<unknown>")


def _get_source_file(obj: Any) -> Path | None:
    try:
        if source_file := inspect.getsourcefile(obj):
            return Path(source_file)
    except TypeError:
        pass
    return None


def _get_source_lineno(obj: Any) -> int:
    try:
        lineno: int
        _lines, lineno = inspect.getsourcelines(obj)
    except (OSError, TypeError):
        pass
    else:
        return lineno
    return 0



import functools
from collections.abc import Callable

from liblaf.grapes import pretty

from . import callback
from ._base import TimerRecords


class TimedFunction[**P, T]:
    timing: TimerRecords
    _fn: Callable[P, T]

    def __init__(self, fn: Callable[P, T], /, *, timing: TimerRecords) -> None:
        self._fn = fn
        self.timing = timing.replace_if_none(
            name=pretty.func(fn).plain or "Function",
            cb_finish=callback.log_summary(depth=3),
            cb_stop=callback.log_record(depth=4),
        )
        functools.update_wrapper(self, fn)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        self.timing.start()
        result: T = self._fn(*args, **kwargs)
        self.timing.stop()
        return result



import contextlib
import enum
import types
from collections.abc import Callable, Iterable, Sequence
from typing import Self, overload, override

import attrs

from liblaf.grapes import utils
from liblaf.grapes.timing import callback

from ._base import Callback, TimerRecords
from ._function import TimedFunction
from ._iterable import TimedIterable


class TimerMode(enum.Enum):
    CONTEXT_MANAGER = enum.auto()
    MANUAL = enum.auto()


@attrs.define
class Timer(
    contextlib.AbstractAsyncContextManager,
    contextlib.AbstractContextManager,
    TimerRecords,
):
    _mode: TimerMode = TimerMode.MANUAL
    _user_name: str | None = attrs.field(default=None, init=False)
    _user_timers: Sequence[str] | None = attrs.field(default=None, init=False)
    _user_cb_finish: Callback | None = attrs.field(default=None, init=False)
    _user_cb_start: Callback | None = attrs.field(default=None, init=False)
    _user_cb_stop: Callback | None = attrs.field(default=None, init=False)

    async def __aenter__(self) -> Self:
        self.mode = TimerMode.CONTEXT_MANAGER
        super().start()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        super().stop()

    @overload
    def __call__[**P, T](self, fn: Callable[P, T], /) -> TimedFunction[P, T]: ...
    @overload
    def __call__[**P, T](
        self, iterable: Iterable[T], /, *, total: int | None = None
    ) -> TimedIterable[T]: ...
    def __call__[**P, T](
        self,
        fn_or_iterable: Callable[P, T] | Iterable[T],
        /,
        *,
        total: int | None = None,
    ) -> TimedFunction[P, T] | TimedIterable[T]:
        timing = TimerRecords(
            name=self._user_name,
            timers=self.timers,
            cb_finish=self._user_cb_finish,
            cb_start=self._user_cb_start,
            cb_stop=self._user_cb_stop,
        )
        if callable(fn_or_iterable):
            return TimedFunction(fn_or_iterable, timing=timing)
        return TimedIterable(fn_or_iterable, timing=timing, total=total)

    def __enter__(self) -> Self:
        self.mode = TimerMode.CONTEXT_MANAGER
        super().start()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        super().stop()

    def __attrs_post_init__(self) -> None:
        self._user_name = self.name
        self._user_cb_finish = self.cb_finish
        self._user_cb_start = self.cb_start
        self._user_cb_stop = self.cb_stop

    @property
    def mode(self) -> TimerMode:
        return self._mode

    @mode.setter
    def mode(self, value: TimerMode) -> None:
        if self._mode == value:
            return
        match value:
            case TimerMode.CONTEXT_MANAGER:
                self.name = self._user_name or "Block"
                self.cb_stop = self._user_cb_stop or callback.log_record(depth=4)
                self.cb_finish = self._user_cb_finish or callback.log_summary(depth=3)
            case TimerMode.MANUAL:
                self.name = self._user_name or "Timer"
                self.cb_stop = self._user_cb_stop or callback.log_record(depth=3)
                self.cb_finish = self._user_cb_finish or callback.log_summary(depth=3)
        self._mode = value

    @override
    def replace_if_none(
        self,
        *,
        name: str | None = None,
        timers: Sequence[str] | None = None,
        cb_finish: Callback | None = None,
        cb_start: Callback | None = None,
        cb_stop: Callback | None = None,
    ) -> Self:
        return attrs.evolve(
            self,
            name=utils.first_not_none(self._user_name, name),
            timers=utils.first_not_none(self._user_timers, timers),
            cb_finish=utils.first_not_none(self._user_cb_finish, cb_finish),
            cb_start=utils.first_not_none(self._user_cb_start, cb_start),
            cb_stop=utils.first_not_none(self._user_cb_stop, cb_stop),
        )

    @override
    def start(self) -> None:
        self.mode = TimerMode.MANUAL
        super().start()


@overload
def timer[**P, T](
    name: str | None = None,
    *,
    timers: Sequence[str] = ("perf",),
    cb_finish: Callback | None = None,
    cb_start: Callback | None = None,
    cb_stop: Callback | None = None,
) -> Timer: ...
@overload
def timer[**P, T](
    fn: Callable[P, T],
    /,
    *,
    name: str | None = None,
    timers: Sequence[str] = ("perf",),
    cb_finish: Callback | None = None,
    cb_start: Callback | None = None,
    cb_stop: Callback | None = None,
) -> TimedFunction[P, T]: ...
@overload
def timer[**P, T](
    iterable: Iterable[T],
    /,
    *,
    name: str | None = None,
    timers: Sequence[str] = ("perf",),
    total: int | None = None,
    cb_finish: Callback | None = None,
    cb_start: Callback | None = None,
    cb_stop: Callback | None = None,
) -> TimedIterable[T]: ...
def timer(*args, **kwargs) -> TimedFunction | TimedIterable | Timer:
    fn_iterable_name: Callable | Iterable | str | None = args[0] if args else None
    if callable(fn_iterable_name):
        return TimedFunction(fn_iterable_name, timing=TimerRecords(**kwargs))
    if isinstance(fn_iterable_name, str):
        kwargs.setdefault("name", fn_iterable_name)
        return Timer(**kwargs)
    if isinstance(fn_iterable_name, Iterable):
        total: int | None = kwargs.pop("total", None)
        return TimedIterable(
            fn_iterable_name, timing=TimerRecords(**kwargs), total=total
        )
    return Timer(**kwargs)



import collections
import math
import statistics
import textwrap
from collections.abc import Callable, Generator, Mapping, Sequence
from typing import Self, overload, override

import attrs
from loguru import logger

from liblaf.grapes import human, utils

from ._get_time import get_time


@attrs.define
class BaseTimer:
    name: str | None = attrs.field(default=None)
    timers: Sequence[str] = attrs.field(default=("perf",))
    _time_start: dict[str, float] = attrs.field(init=False, factory=dict)
    _time_stop: dict[str, float] = attrs.field(init=False, factory=dict)

    def __bool__(self) -> bool:
        return True

    @property
    def default_timer(self) -> str:
        return self.timers[0]

    @property
    def record(self) -> Mapping[str, float]:
        return {timer: self.elapsed(timer) for timer in self.timers}

    def elapsed(self, timer: str | None = None) -> float:
        if timer is None:
            timer = self.default_timer
        if timer in self._time_stop:
            return self._time_stop[timer] - self._time_start[timer]
        return get_time(timer) - self._time_start[timer]

    def start(self) -> None:
        for timer in self.timers:
            self._time_start[timer] = get_time(timer)
        self._time_stop.clear()

    def stop(self) -> None:
        for timer in self.timers:
            self._time_stop[timer] = get_time(timer)


type Callback = Callable[["TimerRecords"], None]


@attrs.define
class TimerRecords(BaseTimer):
    cb_finish: Callable | None = attrs.field(default=None)
    cb_start: Callable | None = attrs.field(default=None)
    cb_stop: Callable | None = attrs.field(default=None)
    _records: dict[str, list[float]] = attrs.field(
        init=False, factory=lambda: collections.defaultdict(list)
    )

    @overload
    def __getitem__(self, key: int) -> Mapping[str, float]: ...
    @overload
    def __getitem__(self, key: str) -> Sequence[float]: ...
    def __getitem__(self, key: int | str) -> Mapping[str, float] | Sequence[float]:
        if isinstance(key, int):
            return self.row(key)
        if isinstance(key, str):
            return self.column(key)
        raise KeyError(key)

    def __len__(self) -> int:
        return len(self._records[self.default_timer])

    @property
    def columns(self) -> Sequence[str]:
        return self.timers

    @property
    def count(self) -> int:
        return self.n_rows

    @property
    def n_columns(self) -> int:
        return len(self.timers)

    @property
    def n_rows(self) -> int:
        return len(self.column())

    def clear(self) -> None:
        self._records.clear()

    def column(self, timer: str | None = None) -> Sequence[float]:
        timer = timer or self.default_timer
        return self._records[timer]

    def human_record(self, index: int = -1) -> str:
        name: str = self.name or "Timer"
        text: str = f"{name} > "
        items: list[str] = []
        for timer, value in self[index].items():
            human_duration: str = human.human_duration(value)
            items.append(f"{timer}: {human_duration}")
        text += ", ".join(items)
        return text

    def human_summary(self, stats: Sequence[str] = ("mean+std", "median")) -> str:
        name: str = self.name or "Timer"
        header: str = f"{name} (total: {self.n_rows})"
        if self.n_rows == 0:
            return header
        body: str = ""
        for timer in self.columns:
            body += f"{timer} > "
            items: list[str] = []
            for stat in stats:
                human_stat_name: str = stat
                human_duration: str
                match stat:
                    case "max":
                        human_duration = human.human_duration(self.max(timer))
                    case "mean":
                        human_duration = human.human_duration(self.mean(timer))
                    case "mean+std":
                        human_stat_name = "mean"
                        human_duration = human.human_duration_with_variance(
                            self.mean(timer), self.std(timer)
                        )
                    case "median":
                        human_duration = human.human_duration(self.median(timer))
                    case "min":
                        human_duration = human.human_duration(self.min(timer))
                    case "std":
                        human_duration = human.human_duration(self.std(timer))
                    case _:
                        msg: str = f"Unknown statistic: {stat}"
                        raise ValueError(msg)
                items.append(f"{human_stat_name}: {human_duration}")
            body += ", ".join(items) + "\n"
        body = body.strip()
        summary: str = header + "\n" + textwrap.indent(body, "    ")
        return summary

    def iter_columns(self) -> Generator[tuple[str, Sequence[float]]]:
        yield from self._records.items()

    def iter_rows(self) -> Generator[Mapping[str, float]]:
        for index in range(self.n_rows):
            yield self.row(index)

    def log_record(
        self,
        depth: int = 1,
        index: int = -1,
        level: int | str = "DEBUG",
        threshold: float = 0.1,  # seconds
    ) -> None:
        if index == -1 and self.elapsed() < threshold:
            return
        logger.opt(depth=depth).log(level, self.human_record(index=index))

    def log_summary(
        self,
        depth: int = 1,
        level: int | str = "INFO",
        stats: Sequence[str] = ("mean+std", "median"),
    ) -> None:
        logger.opt(depth=depth).log(level, self.human_summary(stats=stats))

    def replace_if_none(
        self,
        *,
        name: str | None = None,
        timers: Sequence[str] | None = None,
        cb_finish: Callback | None = None,
        cb_start: Callback | None = None,
        cb_stop: Callback | None = None,
    ) -> Self:
        return attrs.evolve(
            self,
            name=utils.first_not_none(self.name, name),
            timers=utils.first_not_none(self.timers, timers),
            cb_finish=utils.first_not_none(self.cb_finish, cb_finish),
            cb_start=utils.first_not_none(self.cb_start, cb_start),
            cb_stop=utils.first_not_none(self.cb_stop, cb_stop),
        )

    def row(self, index: int) -> Mapping[str, float]:
        return {timer: values[index] for timer, values in self._records.items()}

    # region statistics

    def max(self, timer: str | None = None) -> float:
        return max(self.column(timer))

    def mean(self, timer: str | None = None) -> float:
        return statistics.mean(self.column(timer))

    def median(self, timer: str | None = None) -> float:
        return statistics.median(self.column(timer))

    def min(self, timer: str | None = None) -> float:
        return min(self.column(timer))

    def std(self, timer: str | None = None) -> float:
        column: Sequence[float] = self.column(timer)
        if len(column) < 2:
            return math.nan
        return statistics.stdev(column)

    # endregion statistics

    def _append(
        self, seconds: Mapping[str, float] = {}, nanoseconds: Mapping[str, float] = {}
    ) -> None:
        for key, value in seconds.items():
            self._records[key].append(value)
        for key, value in nanoseconds.items():
            self._records[key].append(value * 1e-9)

    @override
    def start(self) -> None:
        if callable(self.cb_start):
            self.cb_start(self)
        super().start()

    @override
    def stop(self) -> None:
        super().stop()
        self._append(seconds=self.record)
        if callable(self.cb_stop):
            self.cb_stop(self)

    def finish(self) -> None:
        if callable(self.cb_finish):
            self.cb_finish(self)



import functools
from collections.abc import Sequence
from typing import overload

from ._base import Callback, TimerRecords


@overload
def log_record(
    *,
    depth: int = 1,
    index: int = -1,
    level: int | str = "DEBUG",
    threshold: float = 0.1,  # seconds
) -> Callback: ...
@overload
def log_record(
    records: TimerRecords,
    /,
    *,
    depth: int = 1,
    index: int = -1,
    level: int | str = "DEBUG",
    threshold: float = 0.1,  # seconds
) -> None: ...
def log_record(
    records: TimerRecords | None = None,
    /,
    *,
    depth: int = 1,
    index: int = -1,
    level: int | str = "DEBUG",
    threshold: float = 0.1,  # seconds
) -> Callback | None:
    if records is None:
        return functools.partial(
            log_record, depth=depth, index=index, level=level, threshold=threshold
        )  # pyright: ignore[reportReturnType]
    records.log_record(depth=depth, index=index, level=level, threshold=threshold)
    return None


@overload
def log_summary(
    *,
    depth: int = 1,
    level: int | str = "INFO",
    stats: Sequence[str] = ("mean+std", "median"),
) -> Callback: ...
@overload
def log_summary(
    records: TimerRecords,
    /,
    *,
    depth: int = 1,
    level: int | str = "INFO",
    stats: Sequence[str] = ("mean+std", "median"),
) -> None: ...
def log_summary(
    records: TimerRecords | None = None,
    /,
    *,
    depth: int = 1,
    level: int | str = "INFO",
    stats: Sequence[str] = ("mean+std", "median"),
) -> Callback | None:
    if records is None:
        return functools.partial(log_summary, depth=depth, level=level, stats=stats)
    records.log_summary(depth=depth, level=level, stats=stats)
    return None


__all__ = ["log_record", "log_summary"]



from collections.abc import Iterable, Iterator

from . import callback
from ._base import TimerRecords


class TimedIterable[T]:
    timing: TimerRecords
    _iterable: Iterable[T]
    _total: int | None = None

    def __init__(
        self,
        iterable: Iterable[T],
        /,
        *,
        timing: TimerRecords,
        total: int | None = None,
    ) -> None:
        self.timing = timing.replace_if_none(
            name="Iterable",
            cb_stop=callback.log_record(depth=4),
            cb_finish=callback.log_summary(depth=4),
        )
        self._iterable = iterable
        self._total = total

    def __contains__(self, x: object, /) -> bool:
        return x in self._iterable  # pyright: ignore[reportOperatorIssue]

    def __len__(self) -> int:
        if self._total is None:
            return len(self._iterable)  # pyright: ignore[reportArgumentType]
        return self._total

    def __iter__(self) -> Iterator[T]:
        # measure generator + consumer time
        self.timing.start()
        for item in self._iterable:
            yield item
            self.timing.stop()
            self.timing.start()
        self.timing.finish()
